"""
Pytest configuration and shared fixtures for FAIRMeta unit tests.

This module defines reusable fixtures for metadata configuration,
API input data, and a local Radboud FDP client.
"""

import os
from datetime import datetime, timedelta, timezone
from typing import Any, Dict

import isodate
import pytest
from dotenv import load_dotenv

from fairmeta import RadboudFDPClient
from test_utils import merge_nested_dicts

load_dotenv()


# ---------------------------------------------------------------------
# Configuration fixtures
# ---------------------------------------------------------------------

@pytest.fixture(params=["minimal", "full"])
def metadata_config(request: pytest.FixtureRequest) -> Dict[str, Any]:
    """Return a metadata configuration dictionary.

    The fixture is parametrized to return either a minimal or a fully
    populated configuration.
    """
    base_config: Dict[str, Any] = {
        "catalog": {
            "mapping": {
                "challenge_description": ["description"],
                "challenge_title": ["title"],
            },
            "contact_point": {
                "hasEmail": "test@testing.com",
                "fn": "David Tester",
            },
            "publisher": {
                "mbox": "publisher@publishing.com",
                "identifier": ["identification"],
                "name": ["uitgever"],
                "homepage": "https://uitgeverij.nl",
            },
            "license": "cc0",
            "dataset": {
                "mapping": {
                    "archive_description": ["description"],
                    "archive_title": ["title"],
                    "challenge_url": ["identifier"],
                    "challenge_keywords": ["keyword"],
                },
                "access_rights": "non_public",
                "contact_point": {
                    "hasEmail": "support@test.com",
                    "fn": "testing support",
                },
                "creator": {
                    "mbox": "person@testing.com",
                    "identifier": ["test person identifier"],
                    "name": ["datasetmaker"],
                    "homepage": "https://datasetmaker.org",
                },
                "publisher": {
                    "mbox": "datapublisher@publishing.com",
                    "identifier": ["identification data"],
                    "name": ["uitgever data"],
                    "homepage": "https://uitgeverij.nl/data",
                },
                "keyword": ["Test platform"],
                "theme": ["HEAL"],
                "applicable_legislation": "https://www.legislation.com",
                "distribution": {
                    "mapping": {
                        "distribution_access_url": ["access_url"],
                        "distribution_size": ["byte_size"],
                        "distribution_format": ["format"],
                    },
                    "license": "cc0",
                    "rights": (
                        "https://www.example.com/contracts/"
                        "definitely_a_real_DPA.pdf"
                    ),
                },
            },
        }
    }

    if request.param == "full":
        extended_config: Dict[str, Any] = {
            "catalog": {
                "publisher": {
                    "spatial": [
                        "http://publications.europa.eu/resource/authority/country/NLD"
                    ],
                    "publisher_note": "Notitie",
                    "publisher_type": (
                        "http://purl.org/adms/publishertype/"
                        "Academia-ScientificOragnisation"
                    ),
                },
                "applicable_legislation": "https://www.legislation.com",
                "creator": {
                    "mbox": "catalog@testing.com",
                    "identifier": ["catalogtest person identifier"],
                    "name": ["catalogmaker"],
                    "homepage": "https://catalogmaker.org",
                },
                "geographical_coverage": (
                    "https://www.geonames.org/countries/NL/"
                    "the-netherlands.html"
                ),
                "homepage": "https://homepage.org",
                "language": "eng",
                "license": "cc0",
                "modification_date": datetime.now(timezone.utc),
                "release_date": datetime.now(timezone.utc),
                "rights": "https://www.websitewithfreetextrights.com",
                "dataset": {
                    "code_values": "https://www.wikidata.org/wiki/Q32566",
                    "coding_system": "https://www.wikidata.org/wiki/Q81095",
                    "conforms_to": "https://www.wikidata.org/wiki/Q81095",
                    "distribution": {
                        "applicable_legislation": "https://www.legislation.com",
                        "compression_format": (
                            "https://www.iana.org/assignments/"
                            "media-types/application/zip"
                        ),
                        "description": [
                            "Description of the distribution",
                            "Description in another language",
                        ],
                        "documentation": "https://documentation.com",
                        "download_url": "https://google.com",
                        "language": ["eng", "nld"],
                        "media_type": (
                            "https://www.iana.org/assignments/"
                            "media-types/text/csv"
                        ),
                        "modification_date": datetime.now(timezone.utc),
                        "packaging_format": "https://package_information.com",
                        "release_date": datetime.now(timezone.utc),
                        "status": "completed",
                        "temporal_resolution": str(timedelta(days=1)),
                        "title": ["title of distribution"],
                    },
                    "frequency": "daily",
                    "purpose": "https://purpose.com",
                    "geographical_coverage": "https://nijmegen.nl",
                    "is_referenced_by": "https://doi.org",
                    "language": "nld",
                    "legal_basis": "InformedConsent",
                    "maximum_typical_age": 55,
                    "minimum_typical_age": 29,
                    "modification_date": datetime.now(timezone.utc),
                    "number_of_records": 99,
                    "number_of_unique_individuals": 88,
                    "personal_data": "https://w3id.org/dpv/pd#Household",
                    "population_coverage": (
                        "Adults aged 18–65 diagnosed with type 2 diabetes "
                        "in the Netherlands between 2015 and 2020"
                    ),
                    "purpose": "https://w3id.org/dpv#CustomerManagement",
                    "release_date": datetime.now(timezone.utc),
                    "temporal_resolution": isodate.duration_isoformat(
                        timedelta(days=1)
                    ),
                    "type": "https://www.type.nl",
                    "status": "withdrawn",
                    "version": "1",
                    "version_notes": ["changed nothing", "still nothing"],
                    "was_generated_by": "https://me.nl",
                },
            }
        }

        return merge_nested_dicts(base_config, extended_config)

    return base_config


# ---------------------------------------------------------------------
# API data fixtures
# ---------------------------------------------------------------------

@pytest.fixture
def api_data() -> Dict[str, Any]:
    """Return example API data for a single dataset."""
    return {
        "id": "TEST",
        "challenge_description": "Description given by challenge",
        "challenge_title": "Title given by challenge",
        "archive_description": "Description given by archive",
        "archive_title": "Title given by archive",
        "challenge_url": "https://challenge.org/slug",
        "challenge_keywords": ["Medical", "keyword2"],
        "distribution_access_url": "https://testing.com/dist1",
        "distribution_size": 489,
        "distribution_format": (
            "http://publications.europa.eu/resource/authority/file-type/PDF"
        ),
    }

@pytest.fixture
def dataset_expansion_config() -> Dict[str, Any]:
    """Return dataset-specific configuration expansion."""
    return {
        "id": "TEST",
        "dataset": {
            "keyword": ["CT", "Prostate"],
            "maximum_typical_age": 999,
            "minimum_typical_age": 1,
            "distribution": {"byte_size": 69},
        },
    }


@pytest.fixture
def extra_api_data() -> Dict[str, Any]:
    """Return additional API data for multi-dataset tests."""
    return {
        "id": "extra_TEST",
        "challenge_description": "extra Description given by challenge",
        "challenge_title": "extra Title given by challenge",
        "archive_description": "extra Description given by archive",
        "archive_title": "extra Title given by archive",
        "challenge_url": "https://challenge.org/slug/extra",
        "challenge_keywords": ["Medical", "keyword2", "extra"],
        "distribution_access_url": "https://testing.com/dist1/extra",
        "distribution_size": 234567,
        "distribution_format": (
            "http://publications.europa.eu/resource/authority/file-type/PDFextra"
        ),
    }

@pytest.fixture
def extra_dataset_expansion_config() -> Dict[str, Any]:
    """Return expansion for the extra test dataset."""
    return {
        "id": "extra_TEST",
        "dataset": {
            "keyword": ["CT", "Prostate"],
            "maximum_typical_age": 999,
            "minimum_typical_age": 1,
            "distribution": {"byte_size": 69},
        },
    }


# ---------------------------------------------------------------------
# FDP client fixture
# ---------------------------------------------------------------------

@pytest.fixture
def fdp_client() -> RadboudFDPClient:
    """Return a RadboudFDPClient configured for local testing."""
    local_base_url = "http://localhost:8080"
    client = RadboudFDPClient(
        token=os.getenv("local_test_FDP_key")
    )
    client.base_url = local_base_url
    client.post_base_url = local_base_url
    return client    


"""
Pydantic schema definitions for Health-RI-compatible metadata models.

This module defines the structural schema (field names and types) for
FAIRMeta metadata records, aligned with DCAT-AP and Health-RI profiles.

The models defined here are intentionally permissive in typing to allow
progressive normalization and transformation in MetadataRecord.
"""

from __future__ import annotations

from datetime import date, datetime
from typing import List, Optional, Union

from pydantic import (
    AnyHttpUrl,
    AnyUrl,
    AwareDatetime,
    BaseModel,
    NaiveDatetime,
)
from sempyro.dcat import AccessRights
from sempyro.geo import Location
from sempyro.hri_dcat import (
    DatasetStatus,
    DatasetTheme,
    DistributionStatus,
    HRIAgent,
    HRIVCard,
)
from sempyro.prov import Activity
from sempyro.time import PeriodOfTime as SempyroPeriodOfTime


# ---------------------------------------------------------------------------
# Support / value object models
# ---------------------------------------------------------------------------

class Agent(BaseModel):
    mbox: str | AnyUrl
    identifier: str | List[str]
    name: str | List[str]
    homepage: AnyUrl

    spatial: Optional[AnyHttpUrl | Location | List[AnyHttpUrl | Location]] = None
    # Unclear what type is
    type: Optional[AnyHttpUrl] = None
    publisher_type: Optional[AnyUrl] = None
    publisher_note: Optional[str] = None


class Kind(BaseModel):
    hasEmail: str | AnyUrl
    fn: str
    
    hasUrl: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None


class PeriodOfTime(BaseModel):
    startDate: Optional[datetime] = None
    endDate: Optional[datetime] = None


# ---------------------------------------------------------------------------
# Main DCAT / Health-RI models
# ---------------------------------------------------------------------------

# These mandatory fields are 
MANDATORY_DISTRIBUTION_FIELDS = {
    "access_url",
    "byte_size",
    "format",
    "license",
    "rights",
}

class Distribution(BaseModel):
    access_url: Optional[AnyHttpUrl] = None
    byte_size: Optional[int] = None
    format: Optional[str | AnyHttpUrl] = None
    license: Optional[str | AnyHttpUrl] = None
    rights: Optional[AnyHttpUrl] = None

    # access_service: Optional[AnyHttpUrl | DataService] = None
    applicable_legislation: Optional[str | AnyHttpUrl | List[str | AnyHttpUrl]] = None
    # checksum: Optional[Checksum] = None 
    compression_format: Optional[AnyHttpUrl] = None
    description: Optional[str | List[str]] = None
    documentation: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    download_url: Optional[AnyHttpUrl] = None
    language: Optional[str | AnyHttpUrl | List[str | AnyHttpUrl]] = None
    linked_schemas: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    media_type: Optional[AnyHttpUrl] = None
    modification_date: Optional[datetime] = None
    packaging_format: Optional[AnyHttpUrl] = None
    release_date: Optional[datetime] = None
    retention_period: Optional[AnyHttpUrl | PeriodOfTime] = None
    status: Optional[str | AnyHttpUrl | DistributionStatus] = None
    temporal_resolution: Optional[str] = None
    title: Optional[str | List[str]] = "Distribution title" #SHACL

    def is_complete(self) -> bool:
        return all(
            getattr(self, field) is not None
            for field in MANDATORY_DISTRIBUTION_FIELDS
        )


class Dataset(BaseModel):
    title: str | List[str]
    description: str | List[str]
    access_rights: str | AccessRights
    contact_point: Kind | HRIVCard
    creator: Agent | HRIAgent | List[Agent | HRIAgent]
    identifier: str
    keyword: List[str] = []
    publisher: Agent | HRIAgent
    theme: str | DatasetTheme | List[str | DatasetTheme]
    applicable_legislation: AnyHttpUrl | List[AnyHttpUrl]

    analytics: Optional[Distribution | List[Distribution]] = None
    code_values: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    coding_system: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    conforms_to: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    distribution: Optional[Distribution | List[Distribution]] = None
    documentation: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    frequency: Optional[str | AnyHttpUrl] = None
    geographical_coverage: Optional[AnyHttpUrl | Location | List[AnyHttpUrl | Location]] = None
    # has_version: Optional[AnyHttpUrl | Dataset | List[AnyHttpUrl | Dataset]] = None # Should point at other dataset? HRIDataset says it should be url though
    health_theme: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None # HealthDCAT-AP is exploring sources that can be used for this
    # in_series: Optional[DatasetSeries | List[DatasetSeries]] = None
    is_referenced_by: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    language: Optional[str | AnyHttpUrl | List[str | AnyHttpUrl]] = None
    legal_basis: Optional[str | AnyHttpUrl | List[str | AnyHttpUrl]] = None
    maximum_typical_age: Optional[int] = None
    minimum_typical_age: Optional[int] = None
    modification_date: Optional[str | AwareDatetime | NaiveDatetime | date] = None
    number_of_records: Optional[int] = None
    number_of_unique_individuals: Optional[int] = None
    # other_identifier: Optional[Identifier | List[Identifier]] = None
    personal_data: Optional[str | AnyHttpUrl | List[str | AnyHttpUrl]] = None
    population_coverage: Optional[str] = None # Cardinality 0..* but doesn't allow for lists
    purpose: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    # qualified_attribution: Optional[AnyHttpUrl | Attribution | List[AnyHttpUrl | Attribution]] = None
    # qualified_relation: Optional[AnyHttpUrl | Relationship | List[AnyHttpUrl | Relationship]] = None
    # quality_annotation: Optional[AnyHttpUrl | QualityCertificate | List[AnyHttpUrl | QualityCertificate]] = None
    release_date: Optional[str | datetime] = None
    retention_period: Optional[PeriodOfTime | List[PeriodOfTime]] = None
    sample: Optional[AnyHttpUrl | Distribution | List[AnyHttpUrl | Distribution]] = None
    source: Optional[AnyHttpUrl | Dataset | List[AnyHttpUrl | Dataset]] = None
    status: Optional[str | DatasetStatus] = None
    temporal_coverage: Optional[PeriodOfTime | List[PeriodOfTime]] = None
    temporal_resolution: Optional[str] = None
    type: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None # Controlled vocabulary
    version: Optional[str] = None
    version_notes: Optional[str | List[str]] = None
    was_generated_by: Optional[AnyHttpUrl | Activity | List[AnyHttpUrl | Activity]] = None


class Catalog(BaseModel):
    contact_point: Kind | HRIVCard
    dataset: List[Dataset]
    description: str | List[str]
    publisher: Agent | HRIAgent
    title: str | List[str]

    applicable_legislation: Optional[AnyHttpUrl | List[AnyHttpUrl]] = None
    # Not currently supported:
    # catalog: Optional[Catalog | List[Catalog]] = None
    creator: Optional[Agent | HRIAgent | List[Agent | HRIAgent]] = None
    geographical_coverage: Optional[AnyHttpUrl | Location | List[AnyHttpUrl | Location]] = None
    # has_part: Optional[Catalog | List[Catalog]] = None
    homepage: Optional[AnyHttpUrl] = None
    language: Optional[str | AnyHttpUrl | List[str | AnyHttpUrl]] = None
    license: Optional[str | AnyHttpUrl] = None
    modification_date: Optional[str | datetime] = None
    release_date: Optional[str | datetime] = None
    rights: Optional[AnyHttpUrl] = None
    # service: Optional[DataService | List[DataService]] = None
    temporal_coverage: Optional[PeriodOfTime] = None
    # Unclear:
    # themes: Optional[str | List[str]] = None


"""
Controlled vocabulary mappings used by FAIRMeta.

This module defines mappings from human-readable string values to
Health-RI / DCAT controlled vocabulary objects provided by sempyro.

These mappings are primarily used by MetadataRecord to normalize
string inputs into standardized, interoperable values.
"""

from sempyro.dcat import AccessRights
from sempyro.dcat.dcat_dataset import Frequency
from sempyro.hri_dcat import (
    DatasetStatus,
    DatasetTheme,
    DistributionStatus,
    GeonovumLicences,
)

ACCESS_RIGHTS_BY_NAME = {right.name: right for right in AccessRights}

DISTRIBUTION_STATUS_BY_NAME = {status.name: status for status in DistributionStatus}

FREQUENCY_BY_NAME = {frequency.name.lower(): frequency for frequency in Frequency}

LICENSE_BY_NAME = {lic.name: lic.value for lic in GeonovumLicences}

DATASET_STATUS_BY_NAME = {status.name: status for status in DatasetStatus}

DATASET_THEME_BY_NAME = {theme.name: theme for theme in DatasetTheme}

__all__ = [
    "ACCESS_RIGHTS_BY_NAME",
    "DISTRIBUTION_STATUS_BY_NAME",
    "FREQUENCY_BY_NAME",
    "LICENSE_BY_NAME",
    "DATASET_STATUS_BY_NAME",
    "DATASET_THEME_BY_NAME",
]


"""
Utility helpers for FAIRMeta test suite.

This module provides shared helper functions to support metadata
transformation, path resolution, and test configuration manipulation.
"""

from copy import deepcopy
from typing import Any, Dict, List, Tuple

from pydantic import AnyHttpUrl, Field
from rdflib import DCTERMS
from sempyro.hri_dcat import HRICatalog

from fairmeta.metadata_model import MetadataRecord


class FDPCatalog(HRICatalog):
    """Test-specific FDP catalog extension with `is_part_of` relation."""

    is_part_of: List[AnyHttpUrl] = Field(
        description="Link to parent object",
        json_schema_extra={
            "rdf_term": DCTERMS.isPartOf,
            "rdf_type": "uri",
        },
    )


def merge_nested_dicts(
    base: Dict[str, Any],
    extension: Dict[str, Any],
) -> Dict[str, Any]:
    """Recursively merge one nested dictionary into another.

    Values in `extension` override or extend values in `base`.

    Args:
        base: Base dictionary to be extended.
        extension: Dictionary containing new or overriding values.

    Returns:
        The updated base dictionary.
    """
    for key, value in extension.items():
        if (
            key in base
            and isinstance(base[key], dict)
            and isinstance(value, dict)
        ):
            merge_nested_dicts(base[key], value)
        else:
            base[key] = value
    return base


def find_metadata_path(
    metadata_config: Dict[str, Any],
    field_key: str,
) -> List[str]:
    """Find the nested metadata path for a mapped API field.

    Args:
        metadata_config: Metadata configuration dictionary.
        field_key: API field key to resolve.

    Returns:
        A list representing the attribute path in the metadata model.
    """
    for key, value in metadata_config.items():
        if key == "mapping" and field_key in value:
            return [value[field_key][0]]

        if isinstance(value, dict):
            sub_path = find_metadata_path(value, field_key)
            if sub_path:
                return [key] + sub_path

    return []


def resolve_metadata_value(
    root_obj: Any,
    path: List[str],
    source: str,
    metadata_config: Dict[str, Any],
):
    """Resolve a value from either metadata config or API data.

    Args:
        root_obj: Root object to traverse.
        path: Attribute path to resolve.
        source: Either `"metadata_config"` or `"api_data"`.
        metadata_config: Metadata configuration dictionary.

    Returns:
        Resolved value (and path when resolving from API data).

    Raises:
        ValueError: If source is invalid.
    """
    if source == "metadata_config":
        for key in path:
            root_obj = getattr(root_obj, key)
            if isinstance(root_obj, list):
                root_obj = root_obj[0]
        return root_obj

    if source == "api_data":
        api_key = path[0]
        internal_path = find_metadata_path(metadata_config, api_key)

        for key in internal_path[:-1]:
            root_obj = getattr(root_obj, key)
            if isinstance(root_obj, list):
                root_obj = root_obj[0]

        value = getattr(root_obj, internal_path[-1])
        return value, internal_path

    raise ValueError(f"Unknown source: {source}")


def build_adapted_metadata_record(
    target: str | None,
    metadata_config: Dict[str, Any],
    api_data: Dict[str, Any],
    path: List[str] | None,
    value: Any | None,
    extra_metadata_configs: List[Dict[str, Any]] | None = None,
) -> MetadataRecord:
    """Create a MetadataRecord with modified metadata or API data.

    Args:
        target: Target source to modify ("metadata_config", "api_data",
            "multi_conf", or None).
        metadata_config: Metadata configuration dictionary.
        api_data: API input data.
        path: Path to the field to modify.
        value: New value to assign.
        extra_metadata_configs: Optional additional metadata configs.

    Returns:
        A constructed MetadataRecord instance.
    """
    if target == "multi_conf":
        return MetadataRecord.from_sources(
            metadata_config,
            api_data,
            data_config=extra_metadata_configs,
        )

    if target == "metadata_config":
        adapted_config = deepcopy(metadata_config)
        data_to_modify = adapted_config
    elif target == "api_data":
        adapted_api_data = deepcopy(api_data)
        data_to_modify = adapted_api_data
    else:
        return MetadataRecord.from_sources(metadata_config, api_data)

    for key in path[:-1]:
        data_to_modify = data_to_modify[key]

    data_to_modify[path[-1]] = value

    if target == "metadata_config":
        return MetadataRecord.from_sources(adapted_config, api_data)

    return MetadataRecord.from_sources(metadata_config, adapted_api_data)


def is_list_field(
    model: MetadataRecord,
    path: List[str],
) -> bool:
    """Determine whether a model field is annotated as a list.

    Args:
        model: MetadataRecord instance.
        path: Attribute path to inspect.

    Returns:
        True if the field is a list, False otherwise.
    """
    for key in path[:-1]:
        model = getattr(model, key)
        if isinstance(model, list):
            model = model[0]

    field_info = model.__class__.model_fields[path[-1]]
    return "List" in str(field_info.annotation)



"""
FAIRMeta command-line interface.

This module provides the CLI entry point for generating, validating,
and publishing metadata records based on a YAML configuration and
external platform metadata sources.
"""

import argparse
import logging
from pathlib import Path
from typing import List

import yaml
from dotenv import load_dotenv

from fairmeta.gatherers import GrandChallenge
from fairmeta.metadata_model import MetadataRecord
from fairmeta.uploader_radboudfdp import RadboudFDPClient

load_dotenv()

def build_argument_parser() -> argparse.ArgumentParser:
    """Create and configure the CLI argument parser.

    Returns:
        Configured ArgumentParser instance.
    """
    parser = argparse.ArgumentParser(
        description="Generate and publish FAIR metadata records.",
    )

    parser.add_argument(
        "config",
        help="YAML configuration file",
    )
    parser.add_argument(
        "platform",
        help="Platform to fetch metadata from",
    )
    parser.add_argument(
        "-s",
        "--slug",
        help="Unique identifier of dataset",
        action="append",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--extra-dataset",
        help="YAML files with extra dataset-specific fields",
        action="append",
        required=False,
    )
    parser.add_argument(
        "--test",
        action="store_true",
        help="Run in test mode",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable verbose logging",
    )
    # parser.add_argument(
    #     'catalog_name', 
    #     help="Name of catalog in FDP",
    # )

    return parser


def load_yaml_file(path: str | Path) -> dict:
    """Load a YAML file into a dictionary.

    Args:
        path: Path to the YAML file.

    Returns:
        Parsed YAML content as a dictionary.
    """
    with open(path, "r", encoding="utf-8") as file:
        return yaml.safe_load(file)
    
def fetch_platform_metadata(
    platform_key: str,
    slugs: List[str],
):
    """Fetch metadata from the specified external platform.

    Args:
        platform_key: Platform identifier (lowercase).
        slugs: Dataset slugs to fetch.

    Returns:
        Platform-specific API metadata.

    Raises:
        ValueError: If the platform is unsupported.
    """
    match platform_key:
        case "grand_challenge":
            platform_client = GrandChallenge()
            return platform_client.gather_data(slugs)
        case _:
            raise ValueError(
                f"Unsupported platform: {platform_key}"
            )



def main() -> None:
    """CLI entry point for FAIRMeta."""
    parser = build_argument_parser()
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.WARNING,
        format="%(levelname)s: %(message)s",
    )

    logging.info("Loading configuration file")
    platform_key = args.platform.lower()

    config_data = load_yaml_file(args.config)
    platforms_config = config_data.get("platforms", {})

    if platform_key not in platforms_config:
        available = ", ".join(platforms_config.keys())
        raise ValueError(
            f"Platform '{platform_key}' not found in config file. "
            f"Available platforms: {available}"
        )

    platform_config = platforms_config[platform_key]

    logging.info("Fetching data from platform: %s", platform_key)
    api_data = fetch_platform_metadata(
        platform_key=platform_key,
        slugs=args.slug,
    )

    extra_dataset_configs = []
    if args.extra_dataset:
        for dataset_path in args.extra_dataset:
            extra_dataset_configs.append(
                load_yaml_file(dataset_path)
            )

    metadata_record = MetadataRecord.from_sources(
        config=platform_config,
        api_data=api_data,
        data_config=extra_dataset_configs,
    )

    logging.info("Validating relaxed metadata schema")
    metadata_record.validate()

    MetadataRecord.transform_values(metadata_record)

    logging.info("Validating strict metadata schema")
    metadata_record.validate()

    fdp_client = RadboudFDPClient(test=args.test)

    # NOTE:
    # This catalog name does not currently affect the generated TTL output.
    catalog_name = "default_catalog_name"

    fdp_client.create_and_publish(
        metadata_record,
        catalog_name,
    )

    logging.info("Done")


if __name__ == "__main__":
    main()


"""
Tests for FAIRMeta API integrations.

This module contains tests for:
- The Grand Challenge gatherer
- Posting, updating, publishing, and deleting metadata in the Radboud FDP
"""

import random
import string
from typing import Any, Dict, List, Tuple, Type

import pytest
import requests
from gcapi import Client
from httpx import HTTPStatusError, Request, Response
from rdflib import Graph

import test_utils
from fairmeta.gatherers.gather_GC import GrandChallenge
from fairmeta import metadata_model


@pytest.mark.parametrize(
    "slug,status_code,expected_exception",
    [
        ("LUNA16", 200, None),
        ("weird", 404, HTTPStatusError),
    ],
)
def test_grand_challenge_gather_challenge(
    monkeypatch: pytest.MonkeyPatch,
    slug: str,
    status_code: int,
    expected_exception: Type[Exception] | None,
) -> None:
    """Test gathering challenge data from the Grand Challenge API."""

    recorded_calls: List[Tuple[str, Dict[str, Any]]] = []

    def fake_get(self, path: str, **kwargs: Any) -> Response:
        recorded_calls.append((path, kwargs))

        if status_code == 200:
            content = {"name": slug, "pk": 1}
        else:
            content = {"detail": "Not found"}

        return Response(
            status_code=status_code,
            json=content,
            request=Request(
                "GET",
                f"https://grand-challenge.org/api/v1/challenges{path}",
            ),
        )

    monkeypatch.setattr(Client, "get", fake_get)

    # -----------------------------------------------------------------
    # Fake GC client internals
    # -----------------------------------------------------------------

    class FakeArchive:
        def __init__(self) -> None:
            self.pk = 2
            self.slug = slug.lower()

    class FakeArchives:
        def detail(self, slug: str) -> FakeArchive:
            return FakeArchive()

    class FakeImages:
        def __init__(self) -> None:
            self._items = [{"id": "img1"}, {"id": "img2"}]

        def iterate_all(self, params: Dict[str, Any]):
            yield from self._items

        def list(self, params: Dict[str, Any]):
            return self._items

    def fake_init(self) -> None:
        self.client = Client(token="token")
        self.client.archives = FakeArchives()
        self.client.images = FakeImages()

    monkeypatch.setattr(
        "fairmeta.gatherers.gather_GC.GrandChallenge.__init__",
        fake_init,
    )

    platform = GrandChallenge()

    if expected_exception:
        with pytest.raises(expected_exception):
            platform._gather_challenge(slug)
    else:
        challenge_data = platform._gather_challenge(slug)

        assert recorded_calls, "Expected GC API to be called"
        assert recorded_calls[0][0].endswith(
            f"/challenges/{slug}"
        ), f"Unexpected request path: {recorded_calls[0][0]}"
        assert isinstance(challenge_data, dict)
        assert challenge_data.get("name") == slug


def test_fdp_create_publish_and_delete(
    fdp_client,
    metadata_config,
    api_data,
) -> None:
    """Test creating, publishing, and deleting metadata in the FDP."""
    schema = test_utils.build_adapted_metadata_record(
        None,
        metadata_config,
        api_data,
        None,
        None,
    )

    metadata_model.MetadataRecord.transform_values(schema)

    catalog_name = "test_catalog"
    created_urls = fdp_client.create_and_publish(schema, catalog_name)

    for url in created_urls:
        response = requests.get(url)
        assert response.status_code == 200

    # Deleting the catalog should cascade-delete all contained resources
    delete_response = fdp_client.delete(
        created_urls[0],
        confirm=False,
    )
    assert delete_response.status_code == 204


def test_fdp_update_catalog_and_dataset(
    fdp_client,
    metadata_config,
    api_data,
) -> None:
    """Test updating catalog- and dataset-level metadata."""
    catalog_url = (
        "http://localhost:8080/catalog/"
        "953d4e65-6ed7-489d-9745-f8f42811d9df"
    )
    dataset_url = (
        "http://localhost:8080/dataset/"
        "14069853-68a6-496d-85ee-d17cc53bf9ab"
    )

    schema = test_utils.build_adapted_metadata_record(
        None,
        metadata_config,
        api_data,
        None,
        None,
    )

    random_letters = string.ascii_lowercase
    new_catalog_description = "".join(
        random.choice(random_letters) for _ in range(25)
    )
    new_dataset_description = "".join(
        random.choice(random_letters) for _ in range(25)
    )

    schema.catalog.description = new_catalog_description
    schema.catalog.dataset[0].description = new_dataset_description

    metadata_model.MetadataRecord.transform_values(schema)

    fdp_client.update(
        "catalog",
        schema.catalog,
        catalog_url,
        dataset_url,
    )
    catalog_response = fdp_client.get(catalog_url)
    assert new_catalog_description in catalog_response.text

    fdp_client.update(
        "dataset",
        schema.catalog.dataset[0],
        dataset_url,
        catalog_url,
    )
    dataset_response = fdp_client.get(dataset_url)
    assert new_dataset_description in dataset_response.text


"""
GrandChallenge data gatherer.

This module provides the `GrandChallenge` class, which interfaces
with the Grand Challenge API to collect metadata from challenges and
archives.
"""

import logging
import os
from urllib.parse import quote
from typing import Any, Dict, List

import gcapi


class GrandChallenge:
    """Gatherer for Grand Challenge challenges and archives."""

    def __init__(self, token: str = None):
        """Initialize the Grand Challenge client.

        Args:
            token: Optional API token. If not provided, reads from
                the environment variable 'GC_key'.
        """
        self.client = gcapi.Client(token=token or os.getenv("GC_key"))

    def gather_data(self, slugs: List[str]) -> List[Dict[str, Any]]:
        """Gather challenge and archive data for a list of slugs.

        For each slug, this method fetches the archive and challenge data
        from the API, merges them into a single dictionary, and optionally
        counts the number of images in the archive.

        Args:
            slugs: List of challenge/archive slugs to gather.

        Returns:
            A list of dictionaries containing combined metadata for each slug.
        """
        combined_data: List[Dict[str, Any]] = []

        for slug in slugs:
            normalized_slug = quote(slug.strip("/"), safe="")
            combined_dict: Dict[str, Any] = {"id": normalized_slug}

            # Gather archive data
            try:
                archive_data = self._gather_archive(normalized_slug)
                combined_dict.update(
                    {f"archive_{k}": v for k, v in archive_data.items()}
                )
            except Exception as e:
                logging.warning(
                    "Could not gather data from archive '%s': %s", normalized_slug, e
                )
                archive_data = {}

            # Gather challenge data
            try:
                challenge_data = self._gather_challenge(normalized_slug)
                combined_dict.update(
                    {f"challenge_{k}": v for k, v in challenge_data.items()}
                )
            except Exception as e:
                logging.warning(
                    "Could not gather data from challenge '%s': %s",
                    normalized_slug,
                    e,
                )
                challenge_data = {}

            # Count images if archive exists
            archive_pk = archive_data.get("pk")
            if archive_pk:
                try:
                    images_iter = self.client.images.iterate_all(params={"archive": archive_pk})
                    first_image = next(images_iter, None)
                    num_images = len(self.client.images.list(params={"archive": archive_pk}))
                    combined_dict.update({"byte_size": num_images})
                except Exception as e:
                    logging.warning(
                        "Could not determine number of images for archive '%s': %s",
                        normalized_slug,
                        e,
                    )

            combined_data.append(combined_dict)

        return combined_data
    
    def _gather_archive(self, slug: str) -> Dict[str, Any]:
        """Retrieve archive metadata from the Grand Challenge API.

        Args:
            slug: Slug identifier for the archive.

        Returns:
            A dictionary containing the archive metadata.
        """
        archive = self.client.archives.detail(slug=slug.lower()) # .lower() because sometimes the archive slug is the challenge slug but lowercase
        return archive.__dict__

    def _gather_challenge(self, slug: str) -> Dict[str, Any]:
        """Retrieve challenge metadata from the Grand Challenge API.

        Args:
            slug: Slug identifier for the challenge.

        Returns:
            A dictionary containing the challenge metadata.
        """
        response = self.client.get(f"/challenges/{slug}", follow_redirects=True)
        response.raise_for_status()
        return response.json()


"""
Unit tests for the MetadataRecord model.

This module validates schema construction, value normalization,
type enforcement, vocabulary normalization, and transformation to
HRI-compliant models.
"""

import pytest
from pydantic import ValidationError
from rdflib import URIRef
from sempyro.geo import Location
from sempyro.hri_dcat import HRIAgent, HRIVCard, HRIDataset, HRIDistribution

from fairmeta import metadata_model, schema_definitions_hri
from fairmeta.metadata_model import MetadataRecord
from test_utils import (
    FDPCatalog,
    build_adapted_metadata_record,
    is_list_field,
    resolve_metadata_value,
)

@pytest.mark.parametrize(
    "target_source,path,value,expected_exception",
    [
        (None, None, None, None),  # Minimal record
        ("metadata_config", ("catalog", "dataset", "contact_point"), "invalid", AttributeError),
        ("metadata_config", ("catalog", "publisher"), "invalid", AttributeError),
        ("metadata_config", ("catalog", "contact_point", "fn"), "", ValueError),
        ("api_data", ("challenge_title",), [], ValueError),
        ("metadata_config", ("catalog", "dataset", "contact_point", "fn"), 'quoted "string"', None),
        ("metadata_config", ("catalog", "homepage"), "", None),
        ("metadata_config", ("catalog", "contact_point", "hasEmail"), "mailto:email@org.org", None),
    ],
)
def test_metadata_record_creation(
    target_source,
    metadata_config,
    api_data,
    path,
    value,
    expected_exception,
):
    """Test creation and validation of MetadataRecord instances."""
    if expected_exception:
        with pytest.raises(expected_exception) as exc_info:
            schema = build_adapted_metadata_record(
                target_source, metadata_config, api_data, path, value
            )
            schema.validate()
            match target_source:
                case "metadata_config":
                    target = resolve_metadata_value(schema, path[:-1], target_source, metadata_config)
                    target = getattr(target, path[-1])
                case "api_data":
                    target, _ = resolve_metadata_value(schema, path, target_source, metadata_config)
                case _:
                    pass       
            if value:
                assert target == value         
        print(exc_info) 
    else:
        schema = build_adapted_metadata_record(
            target_source, metadata_config, api_data, path, value
        )
        assert isinstance(schema, MetadataRecord)
        combined_keywords = (
            metadata_config["catalog"]["dataset"]["keyword"]
            + api_data["challenge_keywords"]
        )
        assert schema.catalog.dataset[0].keyword == combined_keywords
        match target_source:
            case "metadata_config":
                target = resolve_metadata_value(schema, path[:-1], target_source, metadata_config)
                target = getattr(target, path[-1])
                assert schema.api_data == api_data
            case "api_data":
                target, _ = resolve_metadata_value(schema, path, target_source, metadata_config)
                assert schema.config == metadata_config
            case _:
                assert schema.config == metadata_config
                assert schema.api_data == api_data
        if value:
            assert target == value
        schema.validate()

@pytest.mark.parametrize(
    "target_source,path,value,expected_exception",
    [
        ("metadata_config", ("catalog", "contact_point", "fn"), 123, ValidationError),
        ("metadata_config", ("catalog", "contact_point", "fn"), True, ValidationError),
        ("metadata_config", ("catalog", "contact_point", "fn"), None, ValueError),
        ("metadata_config", ("catalog", "contact_point", "fn"), "Firstname Lastname", None),
        ("api_data", ("archive_title",), 123, ValidationError),
        ("api_data", ("archive_title",), "title", None),
    ],
)
def test_field_type_validation(
    target_source,
    metadata_config,
    api_data,
    path,
    value,
    expected_exception,
):
    """Test enforcement of basic field types."""
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value
    )
    if expected_exception:
        with pytest.raises(expected_exception) as exc_info:
            schema.validate() # Bad error: if type is ambigious (Kind | HRIVCard for example) it prints errors for both
        print(exc_info)
    else:
        schema.validate()

@pytest.mark.parametrize(
    "target_source,path,value,expected_exception,expected_message",
    [
        ("metadata_config", ("catalog", "dataset", "theme"), ["HEAL"], None, None),
        ("metadata_config", ("catalog", "dataset", "theme"), ["INVALID_THEME"], ValueError, f"INVALID_THEME is not supported. Supported values: {', '.join(metadata_model.DATASET_THEME_BY_NAME.keys())}"),
        ("metadata_config", ("catalog", "dataset", "access_rights"), "public", None, None),
        ("metadata_config", ("catalog", "dataset", "access_rights"), "NOT_ALLOWED", ValueError, f"NOT_ALLOWED is not supported. Supported values: {', '.join(metadata_model.ACCESS_RIGHTS_BY_NAME.keys())}"),
        ("metadata_config", ("catalog", "dataset", "theme"), "HEAL", None, None),
        ("metadata_config", ("catalog", "dataset", "theme"), "BAD", ValueError, None),
        ("metadata_config", ("catalog", "language"), "Eng", None, None),
        ("metadata_config", ("catalog", "language"), "En", ValueError, None),
        ("metadata_config", ("catalog", "language"), "https://example.com", ValueError, None),
        ("metadata_config", ("catalog", "language"), "http://publications.europa.eu/resource/authority/language/anything", None, None), # Not correct but shouldn't raise error
        ("metadata_config", ("catalog", "dataset", "frequency"), "quarterly", None, None),
        ("metadata_config", ("catalog", "dataset", "frequency"), "dagelijks", ValueError, None),
        ("metadata_config", ("catalog", "license"), "cc_byncsa_30", None, None),
        ("metadata_config", ("catalog", "license"), "Apache-2.0", ValueError, None),
        ("metadata_config", ("catalog", "dataset", "distribution", "status"), "develoP", None, None),
        ("metadata_config", ("catalog", "dataset", "distribution", "status"), "asfd", ValueError, None)
    ],
)
def test_vocabulary_normalization(
    target_source,
    metadata_config,
    api_data,
    path,
    value,
    expected_exception,
    expected_message,
):
    """Test normalization of controlled vocabulary fields."""
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value
    )
    if expected_exception:
        if expected_message:
            with pytest.raises(expected_exception, match=expected_message):
                MetadataRecord._normalize_vocabularies(schema)
        else:
            with pytest.raises(expected_exception) as exc_info:
                MetadataRecord._normalize_vocabularies(schema) # Prints good error
            print(exc_info)
    else:
        MetadataRecord._normalize_vocabularies(schema)

@pytest.mark.parametrize(
    "target_source,path,value,expected_exception",
    [
        ("metadata_config", ("catalog", "contact_point", "hasEmail"), "name", ValueError),
        ("metadata_config", ("catalog", "contact_point", "hasEmail"), "email@org.com", None),
        ("metadata_config", ("catalog", "dataset", "contact_point", "fn"), "anyone", None),
        ("metadata_config", ("catalog", "dataset", "contact_point", "hasUrl"), ["https://example.com"], None),
        ("metadata_config", ("catalog", "contact_point", "hasUrl"), ["no url"], ValueError),
        ("metadata_config", ("catalog", "contact_point", "hasUrl"), None, None),
        ("metadata_config", ("catalog", "contact_point"), [schema_definitions_hri.Kind(hasEmail="email@email.com", fn="name"), HRIVCard(hasEmail="email@email.com", formatted_name="name")], None),
        ("metadata_config", ("catalog", "contact_point"), [schema_definitions_hri.Kind(hasEmail="email@email.com", fn="name"), HRIVCard(hasEmail="email@email.com", formatted_name="name"), "random"], ValueError)
    ],
)
def test_kind_to_hrivcard_transformation(
    target_source,
    metadata_config,
    api_data,
    path,
    value,
    expected_exception,
):
    """Test conversion from Kind to HRIVCard."""
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value
    )
    if expected_exception:
        with pytest.raises(expected_exception) as exc_info:
            MetadataRecord._kind_to_hrivcard(schema) # Prints good error
        print(exc_info)
    else:
        MetadataRecord._kind_to_hrivcard(schema)
        target = resolve_metadata_value(schema, path[:-1], target_source, metadata_config)
        try:
            assert isinstance(target, HRIVCard)
        except:
            target = getattr(target, path[-1])
            assert all(isinstance(t, HRIVCard) for t in target)

        if path[-1] == "hasUrl" and value is not None:
            if isinstance(target, list):
                for t in target:
                    assert t.contact_page is not None
            else:
                assert target.contact_page is not None

@pytest.mark.parametrize(
    "target_source,path,value,expected_exception",
    [
        ("metadata_config", ("catalog", "publisher", "mbox"), "name", ValueError), # No email in mailbox
        ("metadata_config", ("catalog", "publisher", "homepage"), "no link", ValueError), # No link in homepage
        ("metadata_config", ("catalog", "dataset", "creator", "type"), "typen", ValueError), # No link in type
        ("metadata_config", ("catalog", "dataset", "creator", "type"), "https://typen.com", None),
        ("metadata_config", ("catalog", "dataset", "creator", "spatial"), ["https://Nijmegen.com"], None),
        ("metadata_config", ("catalog", "dataset", "creator", "spatial"), ["Nijmegen"], ValueError), # No link in location
        ("metadata_config", ("catalog", "dataset", "creator", "spatial"), [Location(geometry="https://Nijmegen.com")], None),                                                    
        ("metadata_config", ("catalog", "publisher", "identifier"), ["identification"], None),
        ("metadata_config", ("catalog", "publisher", "publisher_type"), None, None),
        ("metadata_config", ("catalog", "dataset", "publisher", "publisher_type"), "https://publishertype.com", None),
        ("metadata_config", ("catalog", "dataset", "publisher", "publisher_type"), ["https://publishertype.com"], ValueError), # Publisher type in a list when it's not supposed to be
        ("metadata_config", ("catalog", "publisher"), [metadata_model.Agent(mbox="dummy@email.com",identifier=["id"],name=["name"],homepage="https://pagina.nl"), HRIAgent(name=["name"],identifier=["id"],mbox="email@email.com",homepage="https://pagina.nl")], None),
        ("metadata_config", ("catalog", "publisher"), [metadata_model.Agent(mbox="dummy@email.com",identifier=["id"],name=["name"],homepage="https://pagina.nl"), HRIAgent(name=["name"],identifier=["id"],mbox="email@email.com",homepage="https://pagina.nl"), "random"], ValueError), # Not Agent in list
        ],
    ) 
def test_agent_to_hriagent_transformation(
    target_source,
    metadata_config,
    api_data,
    path,
    value,
    expected_exception,
):
    """Test conversion from Agent to HRIAgent."""
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value
    )
    if expected_exception:
        with pytest.raises(expected_exception) as exc_info:
            MetadataRecord._agent_to_hriagent(schema) # Prints good error
        print(exc_info)
    else:
        MetadataRecord._agent_to_hriagent(schema)
        target = resolve_metadata_value(schema, path[:-1], target_source, metadata_config)
        try:
            assert isinstance(target, HRIAgent) 
        except:
            target = getattr(target, path[-1])
            assert all(isinstance(t, HRIAgent) for t in target)

        if path[-1] == "type":
            assert target.type is not None
        elif path[-1] == "spatial":
            assert target.spatial is not None

@pytest.mark.parametrize(
    "target_source,path,value,expected_exception",
    [
        (None, None, None, None),
    ],
)
def test_transformation_hri(
    target_source, 
    metadata_config, 
    api_data, 
    path, 
    value, 
    expected_exception,
):
    """Test transformation from MetadataRecord to HRI-compatible objects."""
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value
    )
    schema.transform_values()
    disallowed_fields = {"distribution", "dataset"}
    filtered_fields = {k: v for k, v in vars(schema.catalog).items() if k not in disallowed_fields and v is not None}
    catalog = FDPCatalog(
        is_part_of=[URIRef("https://test.com")],
        dataset=[],
        **filtered_fields)
    for dataset in schema.catalog.dataset:
        filtered_fields = {k: v for k, v in vars(dataset).items() if k not in disallowed_fields and v is not None}
        hri_dataset = HRIDataset(
            **filtered_fields
        )
        for distribution in dataset.distribution:
            filtered_fields = {k: v for k, v in vars(distribution).items() if k not in disallowed_fields and v is not None}
            hri_distribution = HRIDistribution(
                **filtered_fields
            )            

@pytest.mark.parametrize("target_source,path,value,expected_exception,message",
    [
        ("api_data", (["challenge_title"]), None, ValueError, "Likely put null or null equivalent value in required field"),
        ("api_data", (["challenge_title"]), "title", None, None),
        ("metadata_config", ("catalog", "license"), None, None, None),
        ("metadata_config", ("catalog", "license"), "cc0", None, None),
        ("metadata_config", ("catalog", "dataset", "contact_point", "fn"), None, ValueError, "Likely put null or null equivalent value in required field"),
        ("metadata_config", ("catalog", "contact_point", "fn"), "", ValueError, None),
        ("api_data", (["challenge_title"]), [], ValueError, "Likely put null or null equivalent value in required field"),
        ("metadata_config", ("catalog", "license"), "", None, None)
    ],
)
def test_prune_none_fields(
    target_source, 
    metadata_config, 
    api_data, 
    path, 
    value, 
    expected_exception, 
    message
):
    """Tests if drop_none function removes null-equivalent values and doesn't interfere with validation function"""
    schema = build_adapted_metadata_record(target_source, metadata_config, api_data, path, value)
    if expected_exception:
        if message:
            with pytest.raises(expected_exception, match=message):
                schema.validate()
        else:
            with pytest.raises(expected_exception) as exc_info:
                schema.validate() # Good error
            print(exc_info)
    else:
        schema.validate()
        match target_source:
            case "metadata_config":
                target = resolve_metadata_value(schema, path[:-1], target_source, metadata_config)
                target = getattr(target, path[-1])
            case "api_data":
                target, _ = resolve_metadata_value(schema, path, target_source, metadata_config)
            case _:
                pass       
        if value:
            assert target == value
        else:
            assert target == None

@pytest.mark.parametrize(
    "target_source,path,value,exception,message",
    [
        ("metadata_config", ("catalog", "dataset", "applicable_legislation"), "legislature", None, None),
        ("metadata_config", ("catalog", "dataset", "applicable_legislation"), ["https://license.com"], None, None),
        ("metadata_config", ("catalog", "dataset", "purpose"), None, None, None),
        ("metadata_config", ("catalog", "dataset", "purpose"), "purposefield", None, None),
        ("metadata_config", ("catalog", "dataset", "purpose"), ["purpose field", "purpose_2"], None, None),
        ("api_data", (["challenge_url"]), ["idee"], None, None), # Warning?
        ("api_data", (["challenge_url"]), ["idee2", "illegal_id"], TypeError, "Multiple values found for field 'identifier'")
    ],
)
def test_enforce_list_types(
    target_source, 
    metadata_config, 
    api_data, 
    path, 
    value, 
    exception, 
    message,
):  
    """Tests the function that creates lists where it needs to be, and removes lists where they don't need to be"""  
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value
    )
    if exception:
        if message:
            with pytest.raises(exception, match=message):
                MetadataRecord._enforce_list_types(schema)
        else:
            with pytest.raises(exception) as exc_info:
                MetadataRecord._enforce_list_types(schema) # Prints good error
            print(exc_info)
    else:
        MetadataRecord._enforce_list_types(schema)
        match target_source:
            case "metadata_config":
                target = resolve_metadata_value(schema, path[:-1], target_source, metadata_config)
                target = getattr(target, path[-1])
                list_type = is_list_field(schema, path)
            case "api_data":
                target, internal_path = resolve_metadata_value(schema, path, target_source, metadata_config)
                list_type = is_list_field(schema, internal_path)
            case _:
                pass
        if list_type and value is not None:
            assert isinstance(target, list)
        else:
            assert not isinstance(target, list)

@pytest.mark.parametrize(
    "target_source,path,value",
    [
        ("multi_conf", None, None)
    ],
)
def test_extra_datasets(
    target_source, 
    metadata_config, 
    api_data, 
    path, 
    value, 
    extra_api_data, 
    extra_dataset_expansion_config, 
    dataset_expansion_config,
):
    """Tests if multiple metadata_config files are handled correctly"""
    api_data = [api_data, extra_api_data]
    extra_metadata_config = [extra_dataset_expansion_config, dataset_expansion_config]
    schema = build_adapted_metadata_record(
        target_source, metadata_config, api_data, path, value, extra_metadata_configs=extra_metadata_config
    )
    assert schema.catalog.dataset[0].keyword == ["Test platform", "CT", "Prostate", "Medical", "keyword2"]
    assert schema.catalog.dataset[0].maximum_typical_age == extra_metadata_config[0]["dataset"]["maximum_typical_age"]


"""
Client for interacting with the Radboud FAIR Data Point (FDP).

This module provides functionality to create, update, publish, and delete
DCAT-HRI compliant metadata records in the Radboud FDP.
"""

import logging
import os
from typing import List

import requests
from pydantic import AnyHttpUrl, Field
from rdflib import DCTERMS, URIRef
from sempyro.hri_dcat import HRICatalog, HRIDataset, HRIDistribution
from urllib.parse import urljoin, urlparse, urlunparse

from .metadata_model import MetadataRecord


class FDPCatalog(HRICatalog):
    is_part_of: [AnyHttpUrl] = Field(
        description="Link to parent object", 
        json_schema_extra={
            "rdf_term": DCTERMS.isPartOf, 
            "rdf_type": "uri"
        })


class RadboudFDPClient:
    """Client for publishing metadata to the Radboud FAIR Data Point."""

    BASE_URL = "https://fdp.radboudumc.nl"
    TEST_PATH_PREFIX = "/acc"

    def __init__(self, test: bool = False, token: str | None = None):
        """Initialize the FDP client.

        Args:
            test: Whether to use the acceptance (test) environment.
            token: Optional API token. If not provided, the environment
                variable ``Radboud_FDP_key`` is used.
        """
        self.test = test
        self.api_token = token or os.getenv("Radboud_FDP_key")

        self.base_url = self.BASE_URL
        self.post_base_url = (
            f"{self.BASE_URL}{self.TEST_PATH_PREFIX}"
            if test
            else self.BASE_URL
        )

    def get(self, endpoint: str) -> requests.Response:
        """Perform a GET request against the FDP API.

        Args:
            endpoint: Relative API endpoint.

        Returns:
            The HTTP response.
        """
        url = urljoin(f"{self.post_base_url}/", endpoint)
        response = requests.get(url, headers=self._auth_headers())
        self._check_response(response, action="GET")
        return response

    def create_and_publish(
        self,
        metadata_record: MetadataRecord,
        catalog_name: str,
    ) -> List[str]:
        """Create and publish a catalog, datasets, and distributions.

        Args:
            metadata_record: MetadataRecord containing catalog, datasets,
                and distributions.
            catalog_name: Name used to construct the catalog URI.

        Returns:
            A list of URLs of all created FDP resources.
        """
        created_urls: List[str] = []
        excluded_fields = {"dataset", "distribution"}

        # -------------------------
        # Create catalog
        # -------------------------
        catalog_fields = self._filter_fields(
            metadata_record.catalog,
            excluded_fields,
        )
        catalog = FDPCatalog(
            is_part_of=[URIRef(self.base_url)],
            dataset=[],
            **catalog_fields,
        )

        catalog_graph = catalog.to_graph(
            URIRef(f"{self.post_base_url}/catalog/{catalog_name}")
        )
        catalog_response = self._post_turtle(
            catalog_graph.serialize(format="turtle"),
            "catalog",
        )
        catalog_url = catalog_response.headers["Location"]
        created_urls.append(catalog_url)

        # -------------------------
        # Create datasets
        # -------------------------
        for dataset in metadata_record.catalog.dataset:
            dataset_fields = self._filter_fields(dataset, excluded_fields)
            hri_dataset = HRIDataset(**dataset_fields)

            dataset_graph = hri_dataset.to_graph(
                subject=URIRef(hri_dataset.identifier)
            )
            dataset_graph.add(
                (
                    URIRef(hri_dataset.identifier),
                    DCTERMS.isPartOf,
                    URIRef(catalog_url),
                )
            )

            dataset_response = self._post_turtle(
                dataset_graph.serialize(format="turtle"),
                "dataset",
            )
            dataset_url = dataset_response.headers["Location"]
            created_urls.append(dataset_url)

            # -------------------------
            # Create distributions
            # -------------------------
            for distribution in dataset.distribution or []:
                try:
                    distribution_fields = self._filter_fields(
                        distribution,
                        excluded_fields,
                    )

                    # Temporary SHACL workaround
                    HRIDistribution.model_fields["byte_size"].json_schema_extra[
                        "rdf_type"
                    ] = "xsd:nonNegativeInteger"

                    hri_distribution = HRIDistribution(**distribution_fields)

                    access_url = str(hri_distribution.access_url)
                    distribution_uri = URIRef(
                        f"{hri_dataset.identifier}/distribution/"
                        f"{access_url.split('/')[-1]}"
                    )

                    distribution_graph = hri_distribution.to_graph(
                        subject=distribution_uri
                    )
                    distribution_graph.add(
                        (
                            distribution_uri,
                            DCTERMS.isPartOf,
                            URIRef(dataset_url),
                        )
                    )

                    distribution_response = self._post_turtle(
                        distribution_graph.serialize(format="turtle"),
                        "distribution",
                    )
                    distribution_url = distribution_response.headers["Location"]
                    created_urls.append(distribution_url)

                    self._publish(distribution_url)
                except Exception:
                    logging.exception(
                        "Failed to create distribution for dataset %s",
                        hri_dataset.identifier,
                    )

            self._publish(dataset_url)

        self._publish(catalog_url)
        return created_urls

    def update(
        self,
        target_type: str,
        metadata_record: MetadataRecord,
        resource_url: str,
        parent_url: str,
    ) -> requests.Response:
        """Update an existing FDP resource.

        Args:
            target_type: One of ``catalog`` or ``dataset``.
            metadata_record: Updated metadata record.
            resource_url: URL of the resource to update.
            parent_url: URL of the parent resource.

        Returns:
            The HTTP response.
        """
        excluded_fields = {"dataset", "distribution"}

        if target_type == "catalog":
            fields = self._filter_fields(metadata_record, excluded_fields)
            catalog = FDPCatalog(
                is_part_of=[URIRef(self.base_url)],
                dataset=[parent_url],
                **fields,
            )
            graph = catalog.to_graph(URIRef(resource_url))

        elif target_type == "dataset":
            fields = self._filter_fields(metadata_record, excluded_fields)
            dataset = HRIDataset(**fields)
            graph = dataset.to_graph(subject=URIRef(resource_url))
            graph.add(
                (URIRef(resource_url), DCTERMS.isPartOf, URIRef(parent_url))
            )

        else:
            raise ValueError(f"Invalid target_type: {target_type}")

        response = self._put_turtle(
            graph.serialize(format="turtle"),
            resource_url,
        )
        logging.info(
            "Updated %s (expected 200): %s", target_type, response.status_code
        )
        return response

    def delete(self, resource_url: str, confirm: bool = True) -> requests.Response | None:
        """Delete a resource from the FDP.

        Args:
            resource_url: URL of the resource to delete.
            confirm: Whether to require interactive confirmation.

        Returns:
            The HTTP response, or None if deletion was cancelled.
        """
        if confirm and not self._confirm_deletion(resource_url):
            logging.info("Deletion cancelled for %s", resource_url)
            return None

        response = requests.delete(
            resource_url,
            headers=self._auth_headers(),
        )
        self._check_response(response, action="DELETE")
        logging.info("Deleted %s (expected 204)", resource_url)
        return response

    def _post_turtle(self, turtle_data: str, resource_type: str) -> requests.Response:
        url = f"{self.post_base_url}/{resource_type}"
        response = requests.post(
            url,
            headers=self._turtle_headers(),
            data=turtle_data,
            allow_redirects=True,
        )
        self._check_response(response, action="POST")
        logging.info(f"Posting: {resource_type}, response (should be 201): {response}")        
        return response

    def _put_turtle(self, turtle_data: str, resource_url: str) -> requests.Response:
        resource_url = self._apply_test_prefix(resource_url)
        response = requests.put(
            resource_url,
            headers=self._turtle_headers(accept="text/turtle"),
            data=turtle_data,
        )
        self._check_response(response, action="PUT")
        return response

    def _publish(self, resource_url: str) -> requests.Response:
        resource_url = self._apply_test_prefix(resource_url)
        publish_url = f"{resource_url}/meta/state"

        response = requests.put(
            publish_url,
            headers={
                **self._auth_headers(),
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            json={"current": "PUBLISHED"},
        )
        self._check_response(response, action="PUT")
        logging.info(f"Published, this should be 200: {response}")        
        return response

    def _check_response(
        self,
        response: requests.Response,
        action: str = "request",
    ) -> None:
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            logging.error(
                "%s failed (%s): %s",
                action.upper(),
                response.status_code,
                response.url,
            )
            logging.error("Response headers: %s", response.headers)
            logging.error("Response body:\n%s", response.text)
            raise

    def _apply_test_prefix(self, url: str) -> str:
        if not self.test:
            return url

        parsed = urlparse(url)
        return urlunparse(
            parsed._replace(path=f"{self.TEST_PATH_PREFIX}{parsed.path}")
        )

    def _filter_fields(self, obj, excluded_fields: set[str]) -> dict:
        return {
            key: value
            for key, value in vars(obj).items()
            if key not in excluded_fields and value is not None
        }

    def _auth_headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self.api_token}"}

    def _turtle_headers(self, accept: str = "text/turtle") -> dict[str, str]:
        return {
            **self._auth_headers(),
            "Accept": accept,
            "Content-Type": "text/turtle",
        }

    def _confirm_deletion(self, resource_url: str) -> bool:
        while True:
            user_input = input(
                f"Are you sure you want to DELETE {resource_url}? [yes/no]: "
            ).strip().lower()

            if user_input == "yes":
                return True
            if user_input in ("", "n", "no"):
                return False


from importlib.metadata import version
from .metadata_model import MetadataRecord
from .uploader_radboudfdp import RadboudFDPClient

# __version__ = version("fairmeta")

__all__ = (
    "MetadataRecord",
    "gatherers",
    "RadboudFDPClient"
)

def __dir__() -> "list[str]":
    return list(__all__)


"""
Metadata model construction, transformation, and validation for FAIRMeta.

This module defines the MetadataRecord class, which is responsible for:
- Constructing a metadata schema from configuration files and API data
- Populating Health-RI-compatible DCAT metadata models
- Transforming values into controlled vocabularies
- Validating the resulting metadata prior to FDP upload
"""

from __future__ import annotations

import logging
import warnings
from typing import Any, Dict, List, Optional, Union

from pydantic import AnyHttpUrl, BaseModel, TypeAdapter, ValidationError
from sempyro.hri_dcat import HRIAgent, HRIVCard

from .mappings import *
from .schema_definitions_hri import Agent, Catalog, Dataset, Distribution, Kind

logger = logging.getLogger(__name__)


def _is_valid_http_url(value: str) -> bool:
    """Check whether a string is a valid HTTP URL.

    Args:
        value: String to validate.

    Returns:
        True if the string is a valid HTTP URL, False otherwise.
    """
    try:
        TypeAdapter(AnyHttpUrl).validate_python(value)
        return True
    except ValidationError:
        return False

class MetadataRecord(BaseModel):
    """Top-level metadata record for FAIRMeta.

    This class orchestrates schema construction, transformation to Health-RI
    controlled vocabularies, and validation of metadata prior to FDP upload.
    """

    catalog: Catalog
    config: Optional[Dict[str, Any]] = None
    data_configs: Optional[List[Dict[str, Any]]] = None
    api_data: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None

    @classmethod
    def from_sources(
        cls,
        config: Optional[Dict[str, Any]] = None,
        api_data: Optional[List[Dict[str, Any]]] = None,
        data_config: Optional[List[Dict[str, Any]]] = None,
    ) -> MetadataRecord:
        """Create and populate a metadata schema instance.

        Args:
            config: Global configuration defining default metadata fields.
            api_data: Dataset-level metadata retrieved from an external API.
            data_config: Dataset-specific configuration overrides.

        Returns:
            A populated MetadataRecord instance.
        """
        schema_obj = cls.model_construct(
            config=config,
            api_data=api_data,
            data_configs=data_config,
        )

        if schema_obj.config is not None:
            cls._apply_config_defaults(schema_obj, schema_obj.config)

            if schema_obj.api_data:
                datasets = (
                    schema_obj.api_data
                    if isinstance(schema_obj.api_data, list)
                    else [schema_obj.api_data]
                )

                for dataset in datasets:
                    dataset_id = dataset.get("id")
                    dataset_config = None

                    if schema_obj.data_configs:
                        for cfg in schema_obj.data_configs:
                            if cfg.get("id") == dataset_id:
                                dataset_config = cfg
                                break

                    cls._populate_from_api(
                        schema_obj,
                        dataset,
                        schema_obj.config,
                        dataset_config=dataset_config,
                    )
        
        for dataset in schema_obj.catalog.dataset:
            valid_distributions = []
            for distribution in dataset.distribution:
                if distribution.is_complete():
                    valid_distributions.append(distribution)
                else:
                    warnings.warn(
                        f"Distribution rejected for {dataset.title}"
                    )
                
            dataset.distribution = valid_distributions
                    
        return schema_obj
    
    def transform_values(self) -> None:
        """Transform schema values into Health-RI-compliant representations."""
        self._enforce_list_types(self)
        self._normalize_vocabularies(self)
        self._agent_to_hriagent(self)
        self._kind_to_hrivcard(self)

    def validate(self) -> None:
        """Validate the metadata record.

        Ensures that required fields are present and values conform to their
        declared types.
        """
        # The _prune_none_fields function below is necessary because when
        # validating an HRIVCard or HRIAgent which has optional values that
        # are None, it gives a ValidationError
        cleaned = self._prune_none_fields(self) 
        type(self).model_validate(cleaned, strict=True)
        logger.info("Validation successful")
          
    @staticmethod
    def _apply_config_defaults(schema_obj: BaseModel, config: Dict[str, Any]) -> None:
        """Recursively fill schema fields using configuration defaults.

        Args:
            schema_obj: Metadata schema object being populated.
            config: Configuration dictionary.
        """
        try:
            for key, value in config.items():
                if isinstance(value, list):
                    if key == "keyword":
                        setattr(schema_obj, key, getattr(schema_obj, key) + value)
                    else:
                        setattr(schema_obj, key, value)
                else:
                    match key:
                        case "catalog":
                            if not hasattr(schema_obj, key):
                                setattr(schema_obj, key, Catalog.model_construct())
                            MetadataRecord._apply_config_defaults(
                                getattr(schema_obj, key), value
                            )
                        case "creator" | "publisher":
                            if not getattr(schema_obj, key, None):
                                setattr(schema_obj, key, Agent.model_construct())
                            MetadataRecord._apply_config_defaults(
                                getattr(schema_obj, key), value
                            )
                        case "contact_point":
                            if not getattr(schema_obj, key, None):
                                setattr(schema_obj, key, Kind.model_construct())
                            MetadataRecord._apply_config_defaults(
                                getattr(schema_obj, key), value
                            )
                        case "mapping" | "dataset" | "distribution":
                            continue
                        case _:
                            if value:
                                try:
                                    current = getattr(schema_obj, key)
                                    if current is not None:
                                        warnings.warn(
                                            f"Field value overwritten: {key}: "
                                            f"{current} -> {value}"
                                        )
                                except AttributeError:
                                    pass
                                setattr(schema_obj, key, value)

        except AttributeError as exc:
            raise AttributeError(
                "Invalid structure in creator, publisher, or contact_point fields"
            ) from exc

    @staticmethod
    def _populate_from_api(schema_obj, api_data: dict, config: dict, dataset_config = None):
        """Recursively fills in the fields from the api data"""
        for field, value in config.items():
            match field:
                case "catalog":
                    MetadataRecord._populate_from_api(getattr(schema_obj, field), api_data, value, dataset_config)
                case "dataset":
                    if not hasattr(schema_obj, field):
                        setattr(schema_obj, field, [])

                    new_dataset = Dataset.model_construct()
                    getattr(schema_obj, field).append(new_dataset)
                    MetadataRecord._apply_config_defaults(new_dataset, value)
                    if dataset_config:
                        MetadataRecord._apply_config_defaults(new_dataset, dataset_config["dataset"])
                    MetadataRecord._populate_from_api(new_dataset, api_data, value, dataset_config)

                case "distribution":
                    if not hasattr(schema_obj, field) or getattr(schema_obj, field) == None:
                        setattr(schema_obj, field, [])

                    new_distribution = Distribution.model_construct()
                    getattr(schema_obj, field).append(new_distribution)
                    MetadataRecord._apply_config_defaults(new_distribution, value)
                    if dataset_config:
                        distribution_exists = dataset_config["dataset"].get("distribution")
                        if distribution_exists:
                            MetadataRecord._apply_config_defaults(new_distribution, dataset_config["dataset"]["distribution"])
                    MetadataRecord._populate_from_api(new_distribution, api_data, value, dataset_config)

                case "mapping":
                    if isinstance(value, dict):
                        for api_field, internal_fields in value.items():
                            if api_field in api_data:
                                for internal_field in internal_fields:
                                    if api_data[api_field]:
                                        if internal_field == "keyword" and isinstance(schema_obj.keyword, list):
                                            setattr(schema_obj, internal_field, schema_obj.keyword + api_data[api_field]) # Not using extend here because it changes keyword in config
                                        else:
                                            setattr(schema_obj, internal_field, api_data[api_field])


    @staticmethod
    def _enforce_list_types(schema_obj: BaseModel) -> None:
        """Ensure list-typed fields are lists and scalar fields are not."""
        for field_name, field in schema_obj.model_fields.items():
            value = getattr(schema_obj, field_name)

            if isinstance(value, BaseModel):
                MetadataRecord._enforce_list_types(value)
            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, BaseModel):
                        MetadataRecord._enforce_list_types(v)

            is_list_type = "List" in str(field.annotation)

            if is_list_type and value is not None and not isinstance(value, list):
                setattr(schema_obj, field_name, [value])
            elif not is_list_type and isinstance(value, list):
                if len(value) == 1:
                    setattr(schema_obj, field_name, value[0])
                    warnings.warn(
                        f"List provided for field '{field_name}'",
                        stacklevel=2,
                    )
                else:
                    raise TypeError(
                        f"Multiple values found for field '{field_name}'"
                    )

    @staticmethod
    def _normalize_vocabularies(schema_obj: BaseModel) -> None:
        """Convert string fields to controlled vocabulary URIs."""
        for field_name in schema_obj.model_fields:
            value = getattr(schema_obj, field_name)

            if isinstance(value, BaseModel):
                MetadataRecord._normalize_vocabularies(value)
            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, BaseModel):
                        MetadataRecord._normalize_vocabularies(v)

            dict_backed = {
                "access_rights": ACCESS_RIGHTS_BY_NAME,
                "theme": DATASET_THEME_BY_NAME,
                "license": LICENSE_BY_NAME,
                "status": DATASET_STATUS_BY_NAME,
                "frequency": FREQUENCY_BY_NAME,
            }

            transformer_backed = {
                "format",
                "language",
                "legal_basis",
                "personal_data",
                "purpose",
            }

            if field_name in dict_backed and value:
                vocab = dict_backed[field_name]
                if field_name == "status" and isinstance(schema_obj, Distribution):
                    vocab = DISTRIBUTION_STATUS_BY_NAME

                if isinstance(value, list):
                    setattr(
                        schema_obj,
                        field_name,
                        [MetadataRecord._normalize_vocab_value(v, vocab) for v in value],
                    )
                else:
                    setattr(
                        schema_obj,
                        field_name,
                        MetadataRecord._normalize_vocab_value(value, vocab),
                    )

            elif field_name in transformer_backed and value:
                if isinstance(value, list):
                    setattr(
                        schema_obj,
                        field_name,
                        [
                            MetadataRecord._normalize_vocab_value(v, field_name) for v in value
                        ],
                    )
                else:
                    setattr(
                        schema_obj,
                        field_name,
                        MetadataRecord._normalize_vocab_value(value, field_name),
                    )

    @staticmethod
    def _normalize_vocab_value(value: Any, vocab: Any) -> Any:
        """Convert a value to a controlled vocabulary entry."""
        if not isinstance(value, str): # There is a bug where if there are multiple datasets, _normalize_vocab_value is called multiple times per object.
            return value
        
        if not isinstance(value, str):
            return value

        match vocab:
            case "format":
                return MetadataRecord._format_transformation(value)
            case "language":
                return MetadataRecord._language_transformation(value)
            case "legal_basis":
                return MetadataRecord._legal_basis_transformation(value)
            case "personal_data":
                return MetadataRecord._personal_data_transformation(value)
            case "purpose":
                return MetadataRecord._purpose_transformation(value)
            case _:
                try:
                    return vocab[value.lower()]
                except KeyError as exc:
                    raise ValueError(
                        f"{value} is not supported. "
                        f"Supported values: {', '.join(vocab.keys())}"
                    ) from exc        

    @staticmethod
    def _format_transformation(value: str) -> str:
        if not _is_valid_http_url(value):
            return (
                "http://publications.europa.eu/resource/authority/file-type/"
                f"{value}"
            )
        if "publications.europa.eu/resource/authority/file-type/" not in value:
            raise ValueError(
                "Format must use EU file-type authority URI"
            )
        return value

    @staticmethod
    def _language_transformation(value: str) -> str:
        if not _is_valid_http_url(value):
            match value.lower():
                case "nederlands" | "dutch" | "nld":
                    return (
                        "http://publications.europa.eu/resource/authority/language/NLD"
                    )
                case "english" | "engels" | "eng":
                    return (
                        "http://publications.europa.eu/resource/authority/language/ENG"
                    )
                case _:
                    raise ValueError(
                        "Language must be 'nld', 'eng', or an EU language URI"
                    )
        if "publications.europa.eu/resource/authority/language/" not in value:
            raise ValueError("Invalid language URI")
        return value

    @staticmethod
    def _legal_basis_transformation(value: str) -> str:
        return (
            value
            if _is_valid_http_url(value)
            else f"https://w3id.org/dpv#{value}"
        )

    @staticmethod
    def _personal_data_transformation(value: str) -> str:
        return (
            value
            if _is_valid_http_url(value)
            else f"https://w3id.org/dpv/pd#{value}"
        )

    @staticmethod
    def _purpose_transformation(value: str) -> str:
        return (
            value
            if _is_valid_http_url(value)
            else f"https://w3id.org/dpv#{value}"
        )              
                        
    @staticmethod
    def _agent_to_hriagent(schema_obj: BaseModel) -> None:
        """Convert Agent objects to HRIAgent objects."""
        for field_name in schema_obj.model_fields:
            value = getattr(schema_obj, field_name)

            if isinstance(value, Agent):
                setattr(
                    schema_obj,
                    field_name,
                    MetadataRecord._create_hriagent(value),
                )
            elif isinstance(value, list) and any(isinstance(v, Agent) for v in value):
                new_values = []
                for v in value:
                    if isinstance(v, Agent):
                        new_values.append(
                            MetadataRecord._create_hriagent(v)
                        )
                    elif isinstance(v, HRIAgent):
                        new_values.append(v)
                    elif isinstance(v, BaseModel):
                        MetadataRecord._agent_to_hriagent(v)
                        new_values.append(v)
                    else:
                        raise ValueError("Encountered not Agent or HRIAgent in list")
                setattr(schema_obj, field_name, new_values)
            elif isinstance(value, BaseModel):
                MetadataRecord._agent_to_hriagent(value)
            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, BaseModel):
                        MetadataRecord._agent_to_hriagent(v)

    @staticmethod
    def _create_hriagent(agent: Agent) -> HRIAgent:
        kwargs: Dict[str, Any] = {
            "mbox": agent.mbox,
            "identifier": agent.identifier,
            "name": agent.name,
            "homepage": agent.homepage,
        }
        for attr in (
            "spatial",
            "type",
            "publisher_type",
            "publisher_note",
        ):
            val = getattr(agent, attr)
            if val is not None:
                kwargs[attr] = val
        return HRIAgent(**kwargs)

    @staticmethod
    def _kind_to_hrivcard(schema_obj: BaseModel) -> None:
        """Convert Kind objects to HRIVCard objects."""
        for field_name in schema_obj.model_fields:
            value = getattr(schema_obj, field_name)

            if isinstance(value, Kind):
                setattr(
                    schema_obj,
                    field_name,
                    MetadataRecord._create_hrivcard(value),
                )
            elif isinstance(value, list) and any(isinstance(v, Kind) for v in value):
                new_values = []
                for v in value:
                    if isinstance(v, Kind):
                        new_values.append(
                            MetadataRecord._create_hrivcard(v)
                        )
                    elif isinstance(v, HRIVCard):
                        new_values.append(v)
                    elif isinstance(v, BaseModel):
                        MetadataRecord._kind_to_hrivcard(v)
                        new_values.append(v)
                    else:
                        raise ValueError("Encountered not Kind or VCard in list")               
                setattr(schema_obj, field_name, new_values)
            elif isinstance(value, BaseModel):
                MetadataRecord._kind_to_hrivcard(value)
            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, BaseModel):
                        MetadataRecord._kind_to_hrivcard(v)

    @staticmethod
    def _create_hrivcard(kind: Kind) -> HRIVCard:
        kwargs: Dict[str, Any] = {
            "hasEmail": kind.hasEmail,
            "formatted_name": kind.fn,
        }
        if kind.hasUrl is not None:
            kwargs["contact_page"] = kind.hasUrl
        return HRIVCard(**kwargs)
       
    @staticmethod
    def _prune_none_fields(data: Any) -> Any:
        """Remove None values from optional fields for validation."""
        if isinstance(data, BaseModel):
            result = {}
            for name, field in data.model_fields.items():
                try:
                    value = getattr(data, name)
                except:
                    raise ValueError("Likely put null or null equivalent value in required field")
                if value is not None or field.is_required():
                    result[name] = MetadataRecord._prune_none_fields(value)
            return result
        if isinstance(data, dict):
            return {
                k: MetadataRecord._prune_none_fields(v)
                for k, v in data.items()
                if v is not None
            }
        if isinstance(data, list):
            return [MetadataRecord._prune_none_fields(v) for v in data if v is not None]
        return data